from PyQt4 import QtGui, QtCore
from simulation.flocksimulation import FlockSimulation
from flockdrawer import FlockDrawer
from controlwidget import ControlWidget
import math

class FlockGUI(QtGui.QWidget):
	"""
	FlockGUI is a graphical user interface for a flock simulation.
	"""
	def __init__(self):
		super(FlockGUI, self).__init__()
		self.init_ui()
		self.simulation = FlockSimulation(self,
			self.drawer.size().width(), self.drawer.size().height())
		self.init_slots()
		self.show()

	def init_ui(self):
		"""
		Sets various UI parameters
		"""
		self.setWindowTitle("Flocking")
		# self.setStyleSheet("background-color: white;")

		vbox = QtGui.QVBoxLayout()
		vbox.setSpacing(11)
		# vbox.setContentsMargins(0,0,0,0)
		# TODO: fix: the values don't get updated on start
		control = ControlWidget(self, "Boids", 50, 0, 300, 300)
		control.valueChanged.connect(self.boid_count_changed)
		vbox.addWidget(control)
		control = ControlWidget(self, "Separation", 156.60, 0.0, 300.0, 1000)
		control.valueChanged.connect(self.separation_changed)
		vbox.addWidget(control)
		control = ControlWidget(self, "Alignment", 0.34, 0.0, 5.0, 500)
		control.valueChanged.connect(self.alignment_changed)
		vbox.addWidget(control)
		control = ControlWidget(self, "Cohesion", 4.03, 0.0, 10.0, 500)
		control.valueChanged.connect(self.cohesion_changed)
		vbox.addWidget(control)
		control = ControlWidget(self, "Neighborhood radius", 70.40, 0.0, 400.0, 2000)
		control.valueChanged.connect(self.neighbor_radius_changed)
		vbox.addWidget(control)
		epsilon = 1e-6 # not letting the angle go over pi
		control = ControlWidget(self, "Neighborhood angle", 2.5, 0.0, math.pi-epsilon, 100)
		control.valueChanged.connect(self.neighbor_angle_changed)
		vbox.addWidget(control)

		controlbar = QtGui.QWidget(self)
		controlbar.setLayout(vbox)
		controlbar.setSizePolicy(QtGui.QSizePolicy.Maximum, QtGui.QSizePolicy.Maximum)
		# controlbar.setStyleSheet("background-color: white;")

		self.drawer = FlockDrawer(self)
		hbox = QtGui.QHBoxLayout()
		hbox.setContentsMargins(0,0,0,0)
		hbox.setSpacing(0)
		hbox.setSizeConstraint(QtGui.QLayout.SetMinimumSize)
		hbox.addWidget(controlbar)
		hbox.addSpacing(1)
		hbox.addWidget(self.drawer)
		self.setLayout(hbox)

	def init_slots(self):
		"""
		Establishes slots for thread communication
		"""
		self.connect(self.simulation, QtCore.SIGNAL("boids(PyQt_PyObject)"),
			self.update_boids)

	#########################
	##### Callbacks:

	def resizeEvent(self, e):
		"""
		Called when the window is resized
		"""
		width = self.drawer.size().width()
		height = self.drawer.size().height()
		self.emit(QtCore.SIGNAL("resize(int,int)"), width, height)

	def update_boids(self, boids):
		"""
		A callback that updates boids from backend
		"""
		self.drawer.boids = boids
		self.drawer.update()

	def boid_count_changed(self, count):
		"""
		Called when boid count is changed
		"""
		self.emit(QtCore.SIGNAL("boid_count(int)"), int(count))

	def separation_changed(self, weight):
		"""
		Called when separation behaviour weight is changed
		"""
		self.emit(QtCore.SIGNAL("separation_weight(float)"), weight)

	def alignment_changed(self, weight):
		"""
		Called when alignment behaviour weight is changed
		"""
		self.emit(QtCore.SIGNAL("alignment_weight(float)"), weight)

	def cohesion_changed(self, weight):
		"""
		Called when cohesion behaviour weight is changed
		"""
		self.emit(QtCore.SIGNAL("cohesion_weight(float)"), weight)

	def neighbor_radius_changed(self, radius):
		"""
		Called when neighbor radius is changed
		"""
		self.emit(QtCore.SIGNAL("neighbor_radius(float)"), radius)

	def neighbor_angle_changed(self, angle):
		"""
		Called when neighbor angle is changed
		"""
		self.emit(QtCore.SIGNAL("neighbor_angle(float)"), angle)
