import random
import math
from boid import Boid

class World(object):
	"""
	World is the simulation area and stores every boid
	"""
	def __init__(self, width, height):
		self._boids = []
		self.width = width
		self.height = height
		self.weight_separation = 156.60
		self.weight_alignment = 0.34
		self.weight_cohesion = 4.03
		self.neighbor_radius = 70.40
		self.neighbor_angle = 2.5	# radians

		# fakes (during dev): TODO: remove
		self.set_boid_count(50)

	# TODO: reorder methods and add properties etc
	def resize(self, width, height):
		"""
		Changes the size of the world
		"""
		if width < 0:
			raise ValueError("width must be >= 0")
		if height < 0:
			raise ValueError("height must be >= 0")
		self.width = int(width)
		self.height = int(height)

	def get_neighbors(self, boid):
		"""
		Returns the neighbors of given boid in a list
		"""
		def dot_product(a, b):
			"""
			Dot product for two complex numbers
			"""
			return a.real*b.real + a.imag*b.imag

		neighbors = []
		for other in self.boids:
			if other == boid:
				continue
			difference = other.position - boid.position
			if abs(difference) == 0: # boids may occasionally overlap
				continue
			dot = dot_product(boid.orientation, difference)
			# length of orientation should be 1
			cosine = dot / (abs(boid.orientation) * abs(difference))
			# to recover from floating point rounding errors:
			if cosine < -1.0:
				cosine = -1.0
			elif cosine > 1.0:
				cosine = 1.0
			angle = math.acos(cosine)

			if abs(difference) < self.neighbor_radius and\
					angle < self.neighbor_angle:
				neighbors.append(other)

		return neighbors

		# return [other for other in self.get_boids() if other != boid and\
		# 	abs(other.position - boid.position) < self.neighbor_radius]

	def add_boid(self, boid):
		"""
		Adds the given boid to world
		"""
		self.boids.append(boid)

	def add_random_boid(self):
		"""
		Adds a boid with random velocity and position
		"""
		speed_limit = 100
		random_position = complex(random.randrange(self.width),
			random.randrange(self.height))
		random_velocity = complex(random.randrange(-speed_limit, speed_limit),
			random.randrange(-speed_limit, speed_limit))
		self.add_boid(Boid(random_position, random_velocity))

	def remove_boid(self, boid=None):
		"""
		Removes the given boid. If it's None, some boid is removed
		"""
		if boid in self.boids:
			index = self.boids.index(boid)
			self.boids.pop(index)
			return
		self.boids.pop()

	def set_boid_count(self, number):
		"""
		Sets the count of boid by
		adding random boids if the given number is larger than current size
		and removes some boids if it's smaller
		"""
		while len(self.boids) < number:
			self.add_random_boid()
		while len(self.boids) > number:
			self.remove_boid()

	def update(self, dt):
		"""
		Updates the world according to the given time step
		"""
		self.move_all(dt)

	def move_all(self, dt):
		"""
		Moves all the boids in the world according to the given time step
		and applies a force defined by behaviours
		dt is in seconds
		"""
		for boid in self.boids:
			force = self.calc_combined_force(boid)
			boid.move(dt, force)
			self.wrap_position(boid)

	def wrap_position(self, boid):
		"""
		Moves the boid back inside the screen if it's outside
		"""
		boid.position = complex(boid.position.real % self.width,
			boid.position.imag % self.height)

	def reset_weights(self):
		"""
		Resets the behaviour weights to defaults
		"""
		self.weight_separation = 6.75
		self.weight_alignment = 0.17
		self.weight_cohesion = 0.03

	def calc_combined_force(self, boid):
		"""
		Returns a combination of behaviour forces for given boids
		"""
		neighbors = self.get_neighbors(boid)
		if len(neighbors) == 0: # no neighbors, no force
			return complex(0, 0)

		force = self.weight_alignment * self.calc_alignment_force(boid, neighbors)
		force += self.weight_separation * self.calc_separation_force(boid, neighbors)
		force += self.weight_cohesion * self.calc_cohesion_force(boid, neighbors)
		return force

	def calc_alignment_force(self, boid, neighbors):
		force = complex(0, 0)
		for other in neighbors:
			force += other.velocity - boid.velocity
		force = force/len(neighbors)
		return force

	def calc_separation_force(self, boid, neighbors):
		force = complex(0, 0)
		for other in neighbors:
			difference = boid.position - other.position	# repulsive
			distance = abs(difference)
			distance *= distance
			if distance != 0:
				force += 15.0*difference/distance
		return force

	def calc_cohesion_force(self, boid, neighbors):
		avg_position = complex(0, 0)
		for other in neighbors:
			avg_position += other.position
		avg_position = avg_position/len(neighbors)
		force = avg_position - boid.position
		return force

	#######################
	#### Getters & setters:

	def get_width(self):
		return self._width

	def set_width(self, width):
		if width < 0:
			raise ValueError("World width should be >= 0")
		self._width = width
	width = property(get_width, set_width)

	def get_height(self):
		return self._height

	def set_height(self, height):
		if height < 0:
			raise ValueError("World height should be >= 0")
		self._height = height
	height = property(get_height, set_height)

	def get_boids(self):
		"""
		Returns the boids of the world
		"""
		return self._boids
	boids = property(get_boids)

	def get_neighbor_radius(self):
		return self._neighbor_radius

	def set_neighbor_radius(self, radius):
		if radius < 0:
			raise ValueError("Neighbor radius should be >= 0")
		self._neighbor_radius = radius
	neighbor_radius = property(get_neighbor_radius, set_neighbor_radius)

	def get_neighbor_angle(self):
		return self._neighbor_angle

	def set_neighbor_angle(self, angle):
		if angle < 0 or angle > math.pi:
			raise ValueError("Neighbor angle should be between 0 and pi")
		self._neighbor_angle = angle
	neighbor_angle = property(get_neighbor_angle, set_neighbor_angle)

	def get_separation_weight(self):
		return self._weight_separation

	def set_separation_weight(self, weight):
		if weight < 0:
			raise ValueError("Separation weight should be >= 0")
		self._weight_separation = weight
	weight_separation = property(get_separation_weight, set_separation_weight)

	def get_alignment_weight(self):
		return self._weight_alignment

	def set_alignment_weight(self, weight):
		if weight < 0:
			raise ValueError("Alignment weight should be >= 0")
		self._weight_alignment = weight
	weight_alignment = property(get_alignment_weight, set_alignment_weight)

	def get_cohesion_weight(self):
		return self._weight_cohesion

	def set_cohesion_weight(self, weight):
		if weight < 0:
			raise ValueError("Cohesion weight should be >= 0")
		self._weight_cohesion = weight
	weight_cohesion = property(get_cohesion_weight, set_cohesion_weight)
